import logging
import re
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass
from operator import attrgetter
from typing import Any, Iterable, List, Optional, Union

import pandas as pd
from sagemaker.workflow.pipeline import Pipeline


@dataclass
class ValidationResult:
    """Validation result dataclass."""

    validation_name: str
    success: bool
    negative: bool
    message: str
    subject: str


class Rule(ABC):
    """Rule abstract base class."""

    def __init__(self, name: str, negative: bool = False) -> None:
        """Initialize a Rule object.

        :param name: name of the rule
        :type name: str
        :param negative: whether the rule should be inverted, i.e. "not" (default: False)
        :type negative: bool
        :return:
        :rtype:
        """
        self.name: str = name
        self.negative: bool = negative

    @abstractmethod
    def run(self, observed: Any, expected: Any, validation_name: str) -> ValidationResult:
        """Run the rule.

        :param observed: observed value
        :type observed: Any
        :param expected: expected value
        :type expected: Any
        :param validation_name: name of the validation
        :type validation_name: str
        :return: validation result
        :rtype: ValidationResult
        """
        raise NotImplementedError


class Validation(ABC):
    """Abstract class for validations."""

    def __init__(
        self, name: str, paths: Optional[List[str]] = None, rule: Optional[Rule] = None
    ) -> None:
        """Initialize a Validation object.

        :param paths: list of paths to the attributes to be validated
        :type paths: List[str]
        :param name: name of the validation
        :type name: str
        :param rule: rule to be applied to the validation, defaults to None
        :type rule: Optional[Rule]
        :return:
        :rtype:
        """
        self.paths: List[Optional[str]] = paths if paths else []
        self.name: str = name
        self.rule: Rule = rule

    @staticmethod
    def get_filtered_attributes(filter_subject: Iterable[object], path: str) -> List[object]:
        """Get filtered attributes from path.

        :param filter_subject: subject to be filtered
        :type filter_subject: Iterable[object]
        :param path: path to the attribute to be filtered
        :type path: str
        :return: filtered attributes
        :rtype: List[object]
        """
        # TODO: refactor
        filtered_steps = []
        filter_conditions = (
            re.search("\[(.*?)\]", path).group(1).replace(" ", "").split("&&")  # noqa: W605
        )
        filter_conditions = [
            condition.replace("/", ".") for condition in filter_conditions if condition
        ]
        for subject in filter_subject:
            match = []
            for condition in filter_conditions:
                filter_key, filter_value = condition.split("==")
                if attrgetter(filter_key)(subject) != filter_value:
                    match.append(False)
                    continue
                match.append(True)
            if all(match):
                filtered_steps.append(subject)
        return filtered_steps

    @staticmethod
    def get_attribute(
        sagemaker_pipeline: Pipeline,
        paths: List[str],
    ) -> List:
        """Get attribute from pipeline.

        :param sagemaker_pipeline: sagemaker pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :param paths: list of paths to the attributes to be validated
        :type paths: List[str]
        :return: attribute
        :rtype: List
        """
        # TODO: refactor
        result = []
        for path in paths:
            attr_path = path.split(".")[1:]
            sm_pipeline_copy = copy(sagemaker_pipeline)
            for ix, attr in enumerate(attr_path):
                if attr.endswith("]"):
                    has_filter_dict = attr[-2] != "["
                    raw_attr = attr.split("[")[0]
                    sm_pipeline_copy = getattr(sm_pipeline_copy, raw_attr)
                    if has_filter_dict:
                        sm_pipeline_copy = Validation.get_filtered_attributes(
                            sm_pipeline_copy, ".".join(attr_path[ix:])
                        )
                else:
                    sm_pipeline_copy = [
                        getattr(sub_attr, attr)
                        for sub_attr in sm_pipeline_copy
                        if hasattr(sub_attr, attr)
                    ]
            result.append(sm_pipeline_copy)
        return [x for y in result for x in y]

    @abstractmethod
    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Run validation."""
        raise NotImplementedError


class Report:
    """Report class."""

    def __init__(self, results: List[ValidationResult]) -> None:
        """Initialize a Report object.

        :param results: list of validation results
        :type results: List[Dict[str, ValidationResult]]
        :return: None
        :rtype: None
        """
        self.results: List[ValidationResult] = results

    def to_df(self) -> pd.DataFrame:
        """Convert report to pandas DataFrame.

        :return: report as pandas DataFrame
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_records(
            data=[x.__dict__ for x in self.results],
            columns=ValidationResult.__annotations__.keys(),
        )
        return df.reset_index(drop=True)


class ValidationFailedError(Exception):
    """Validation exception class."""

    def __init__(self, validation_result: ValidationResult) -> None:
        """Initialize a ValidationFailedError object.

        :param validation: error message
        :type validation: Validation
        :return: None
        :rtype: None
        """
        self.validation_result = validation_result
        self.message = f"Validation failed: {validation_result.__dict__}"
        super().__init__(self.message)


class Configuration:
    """Configuration class."""

    def __init__(self, validations: List[Validation], sagemaker_pipeline: Pipeline) -> None:
        """Initialize a Configuration object.

        :param validations: List of validations.
        :type validations: List[Validation]
        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :raises ValueError: If validations is empty.
        :raises TypeError: If validations is not a list.
        :raises TypeError: If any validation is not of type Validation.
        :return: None
        :rtype: None
        """
        Configuration._validate_input_validations(validations)
        self.validations: List[Validation] = validations
        self.sagemaker_pipeline: Pipeline = sagemaker_pipeline

    @staticmethod
    def _validate_input_validations(validations: List[Validation]) -> None:
        """Validate input validations.

        :param validations: List of validations.
        :type validations: List[Validation]
        :raises ValueError: If validations is empty.
        :raises TypeError: If validations is not a list.
        :raises TypeError: If any validation is not of type Validation.
        :return: None
        :rtype: None
        """
        if not validations:
            raise ValueError("Validations cannot be empty.")
        if not isinstance(validations, list):
            raise TypeError("Validations must be a list.")
        if not all(isinstance(v, Validation) for v in validations):
            raise TypeError("All validations must be of type Validation.")

    @staticmethod
    def _make_report(
        results: List[ValidationResult], return_df: bool = False
    ) -> Union[Report, pd.DataFrame]:
        """Make a report from a list of results.

        :param results: List of results.
        :type results: List[ValidationResult
        :param return_df: If True, return a pandas.DataFrame instead of a Report object.
        :type return_df: bool
        :return: Report object or pd.DataFrame.
        :rtype: Report or pd.DataFrame
        """
        report = Report(results)
        if return_df:
            return report.to_df()
        return report

    @staticmethod
    def _handle_empty_results(
        result: Optional[ValidationResult], validation: Validation
    ) -> ValidationResult:
        """Handle empty results. If a Validation does not return any results
        (e.g. when no observation were made), a warning is logged and a
        ValidationResult indicating this is added to the result dict.

        :param result: validation_result.
        :type result: Optional[ValidationResult]
        :param validation: Validation object.
        :type validation: Validation
        :return: validation result.
        :rtype: ValidationResult
        """
        if not result:
            logging.warning(
                f"Validation {validation.name} did not return any results. "
                f"Please check if the paths {validation.paths} are correct."
            )
            return ValidationResult(
                validation_name=validation.name,
                success=False,
                message=f"Validation {validation.name} did not return any results. "
                f"Please check if the paths {validation.paths} are correct.",
                subject=validation.paths,
                negative=False,
            )
        return result

    def run(self, fail_fast: bool = False, return_df: bool = False) -> Union[Report, dict]:
        """Run all validations and return a report.

        :param fail_fast: If True, stop validation after the first failure.
        :type fail_fast: bool
        :param return_df: If True, return a pandas dataframe instead of a Report object.
        :type return_df: bool
        :raises ValidationFailedError: If fail_fast is True and a validation fails.
        :return: Report object or pandas dataframe.
        :rtype: Report or dict
        """
        results = []
        for ix, validation in enumerate(self.validations):
            result = validation.run(self.sagemaker_pipeline)
            result = Configuration._handle_empty_results(result, validation)
            results.append(result)
            if not result.success and fail_fast and not (ix == len(self.validations) - 1):
                logging.info(
                    "Validation failed and fail_fast is set to True. Stopping validation "
                    "prematurely."
                )
                raise ValidationFailedError(result)
        return self._make_report(results, return_df)



import pytest
from moto import mock_ecr, mock_iam, mock_lambda, mock_sqs
from sagemaker.inputs import FileSystemInput, TrainingInput
from sagemaker.processing import NetworkConfig, ProcessingInput, ProcessingOutput
from sagemaker.workflow.functions import Join
from sagemaker.workflow.parameters import ParameterString
from sagemaker.workflow.pipeline import ExecutionVariables

from sagemaker_rightline.model import Validation, ValidationResult
from sagemaker_rightline.rules import Contains, Equals
from sagemaker_rightline.validations import (
    ContainerImage,
    PipelineParametersAsExpected,
    StepCallbackSqsQueueExists,
    StepImagesExist,
    StepInputsAsExpected,
    StepKmsKeyIdAsExpected,
    StepLambdaFunctionExists,
    StepNetworkConfigAsExpected,
    StepOutputsAsExpected,
    StepOutputsMatchInputsAsExpected,
    StepRoleNameAsExpected,
    StepRoleNameExists,
    StepTagsAsExpected,
)
from tests.fixtures.constants import (
    TEST_ACCOUNT_ID,
    TEST_LAMBDA_FUNC_NAME,
    TEST_REGION_NAME,
    TEST_ROLE_NAME,
    TEST_SQS_QUEUE_NAME,
    TEST_SQS_QUEUE_URL_BASE,
)
from tests.fixtures.image_details import (
    IMAGE_1_REPOSITORY_NAME,
    IMAGE_1_TAG,
    IMAGE_1_URI,
    IMAGE_2_URI,
)
from tests.fixtures.pipeline import DUMMY_BUCKET, get_sagemaker_pipeline
from tests.utils import (
    create_iam_role,
    create_image,
    create_lambda_function,
    create_sqs_queue,
    ecr_client,
    iam_client,
    lambda_client,
    sqs_client,
)


@pytest.fixture
def sagemaker_pipeline():
    return get_sagemaker_pipeline()


def test_container_image() -> None:
    container_image = ContainerImage(uri=IMAGE_1_URI)
    assert container_image.uri == IMAGE_1_URI
    assert container_image.account_id == TEST_ACCOUNT_ID
    assert container_image.repository == IMAGE_1_REPOSITORY_NAME
    assert container_image.region == TEST_REGION_NAME
    assert container_image.tag == IMAGE_1_TAG


def test_validation_result() -> None:
    success = True
    negative = True
    message = "test-message"
    subject = "test-subject"
    validation_name = "test"
    vr = ValidationResult(
        success=success,
        negative=negative,
        message=message,
        subject=subject,
        validation_name=validation_name,
    )
    assert vr.success == success
    assert vr.negative == negative
    assert vr.message == message
    assert vr.subject == subject
    assert vr.validation_name == validation_name


@mock_ecr
def test_step_image_exists_run_positive(ecr_client, sagemaker_pipeline) -> None:
    container_images = [
        ContainerImage(uri=IMAGE_1_URI),
        ContainerImage(uri=IMAGE_2_URI),
    ]
    with create_image(ecr_client, container_images):
        image_exists = StepImagesExist()
        result = image_exists.run(sagemaker_pipeline)

    assert result.success
    assert not result.negative
    assert result.message.endswith(" exist.")


@mock_ecr
def test_step_image_exists_run_negative(sagemaker_pipeline) -> None:
    image_exists = StepImagesExist()
    result = image_exists.run(sagemaker_pipeline)

    assert not result.success
    assert result.message.endswith(" not exist.")


def test_step_image_exists_wrong_client() -> None:
    with pytest.raises(ValueError):
        _ = StepImagesExist(boto3_client="not-a-boto3-client")


def test_step_lambda_function_exists_wrong_client() -> None:
    with pytest.raises(ValueError):
        _ = StepLambdaFunctionExists(boto3_client="not-a-boto3-client")


def test_step_sqs_queue_exists_wrong_client() -> None:
    with pytest.raises(ValueError):
        _ = StepCallbackSqsQueueExists(boto3_client="not-a-boto3-client")


def test_step_role_name_exists_wrong_client() -> None:
    with pytest.raises(ValueError):
        _ = StepRoleNameExists(boto3_client="not-a-boto3-client")


@pytest.mark.parametrize(
    "parameters_expected,success",
    [
        [
            [
                ParameterString(name="parameter-1", default_value="some-value-1"),
                ParameterString(name="parameter-2", default_value="some-value-2"),
            ],
            True,
        ],
        [[ParameterString(name="parameter-1", default_value="some-value-1")], False],
        [
            [
                ParameterString(name="parameter-1", default_value="some-value-1"),
                ParameterString(name="nonexistent-parameter", default_value="some-value-2"),
            ],
            False,
        ],
        [
            [
                ParameterString(name="parameter-1"),
                ParameterString(name="parameter-2", default_value="some-value-2"),
            ],
            False,
        ],
    ],
)
def test_pipeline_parameters_equals(parameters_expected, success, sagemaker_pipeline) -> None:
    pipeline_parameters = PipelineParametersAsExpected(
        parameters_expected=parameters_expected,
        rule=Equals(),
    )
    result = pipeline_parameters.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "parameters_expected,ignore_default_value,success",
    [
        [
            [
                ParameterString(name="parameter-1"),
                ParameterString(name="parameter-2"),
            ],
            True,
            True,
        ],
        [
            [
                ParameterString(name="parameter-1", default_value="some-value-1"),
                ParameterString(name="parameter-2", default_value="different-default-value"),
            ],
            False,
            False,
        ],
    ],
)
def test_pipeline_parameters_ignore_default_value(
    parameters_expected, ignore_default_value, success, sagemaker_pipeline
) -> None:
    pipeline_parameters = PipelineParametersAsExpected(
        parameters_expected=parameters_expected,
        ignore_default_value=ignore_default_value,
        rule=Equals(),
    )
    result = pipeline_parameters.run(sagemaker_pipeline)
    assert result.success == success


def test_has_parameters_raise() -> None:
    with pytest.raises(ValueError):
        _ = PipelineParametersAsExpected(parameters_expected=[], rule=Equals())


@pytest.mark.parametrize(
    "kms_key_id_expected,success",
    [
        ["some/kms-key-alias", True],
        ["does-not-exist", False],
    ],
)
def test_step_kms_key_id(kms_key_id_expected, success, sagemaker_pipeline) -> None:
    has_kms_key_id_in_processing_output = StepKmsKeyIdAsExpected(
        kms_key_id_expected=kms_key_id_expected,
        rule=Equals(),
    )
    result = has_kms_key_id_in_processing_output.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "kms_key_id_expected,success",
    [
        ["some/kms-key-alias", True],
        ["does-not-exist", False],
    ],
)
def test_step_kms_key_id_filter(kms_key_id_expected, success, sagemaker_pipeline) -> None:
    has_kms_key_id_in_processing_output = StepKmsKeyIdAsExpected(
        kms_key_id_expected=kms_key_id_expected,
        step_name="sm_processing_step_sklearn",
        rule=Equals(),
    )
    result = has_kms_key_id_in_processing_output.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "kms_key_id_expected,success",
    [
        ["some/kms-key-alias", True],
        ["does-not-exist", False],
    ],
)
def test_step_kms_key_id_no_filter(kms_key_id_expected, success, sagemaker_pipeline) -> None:
    has_kms_key_id_in_processing_output = StepKmsKeyIdAsExpected(
        kms_key_id_expected=kms_key_id_expected,
        rule=Equals(),
    )
    result = has_kms_key_id_in_processing_output.run(sagemaker_pipeline)
    assert result.success == success


def test_get_filtered_attributes_steps(sagemaker_pipeline) -> None:
    steps = Validation.get_filtered_attributes(
        filter_subject=sagemaker_pipeline.steps,
        path=".steps[name==sm_processing_step_sklearn && step_type.value==Processing].kms_key",
    )
    assert len(steps) == 1
    assert steps[0].name == "sm_processing_step_sklearn"


def test_step_network_config(
    sagemaker_pipeline,
) -> None:
    network_config_expected = NetworkConfig(
        enable_network_isolation=True,
        security_group_ids=["sg-12345"],
        subnets=["subnet-12345"],
        encrypt_inter_container_traffic=True,
    )
    step_network_config = StepNetworkConfigAsExpected(
        network_config_expected=network_config_expected,
        rule=Equals(),
    )
    result = step_network_config.run(sagemaker_pipeline)
    assert result.success


def test_step_network_config_none_observed(
    sagemaker_pipeline,
) -> None:
    network_config_expected = NetworkConfig(
        enable_network_isolation=True,
        security_group_ids=["sg-12345"],
        subnets=["subnet-12345"],
        encrypt_inter_container_traffic=True,
    )
    step_network_config = StepNetworkConfigAsExpected(
        network_config_expected=network_config_expected,
        rule=Equals(),
    )
    result = step_network_config.rule.run([None], network_config_expected, step_network_config.name)
    assert not result.success


@pytest.mark.parametrize(
    "network_config_expected,success,step_name",
    [
        [
            NetworkConfig(
                enable_network_isolation=True,
                security_group_ids=["sg-12345"],
                subnets=["subnet-12345"],
                encrypt_inter_container_traffic=True,
            ),
            True,
            "sm_processing_step_sklearn",
        ],
        [
            NetworkConfig(
                enable_network_isolation=True,
                security_group_ids=["sg-12345"],
                subnets=["other-subnet"],
                encrypt_inter_container_traffic=False,
            ),
            False,
            "sm_processing_step_sklearn",
        ],
        [NetworkConfig(), False, "sm_processing_step_sklearn"],
        [None, False, "sm_processing_step_sklearn"],
        [NetworkConfig(), False, "sm_processing_step_spark"],
        [
            NetworkConfig(
                enable_network_isolation=True,
                security_group_ids=["sg-12345"],
                subnets=["subnet-12345"],
                encrypt_inter_container_traffic=True,
            ),
            True,
            "sm_training_step_sklearn",
        ],
    ],
)
def test_step_network_config_filter(
    network_config_expected, success, step_name, sagemaker_pipeline
) -> None:
    step_network_config = StepNetworkConfigAsExpected(
        network_config_expected=network_config_expected,
        rule=Equals(),
        step_name=step_name,
    )
    result = step_network_config.run(sagemaker_pipeline)
    assert result.success == success


@mock_iam
@mock_lambda
def test_lambda_function_exists_positive(lambda_client, iam_client, sagemaker_pipeline) -> None:
    with create_lambda_function(lambda_client, iam_client, [TEST_LAMBDA_FUNC_NAME]):
        lambda_function_exists = StepLambdaFunctionExists()
        result = lambda_function_exists.run(sagemaker_pipeline)
    assert result.success


@mock_lambda
def test_lambda_function_exists_negative(lambda_client, sagemaker_pipeline) -> None:
    lambda_function_exists = StepLambdaFunctionExists()
    result = lambda_function_exists.run(sagemaker_pipeline)
    assert not result.success


@mock_iam
def test_role_exists_positive(iam_client, sagemaker_pipeline) -> None:
    with create_iam_role(iam_client, [TEST_ROLE_NAME]):
        role_exists = StepRoleNameExists()
        result = role_exists.run(sagemaker_pipeline)
    assert result.success


@mock_iam
def test_role_exists_negative(iam_client, sagemaker_pipeline) -> None:
    role_exists = StepRoleNameExists()
    result = role_exists.run(sagemaker_pipeline)
    assert not result.success


@pytest.mark.parametrize(
    "role_name_expected,success",
    [
        [TEST_ROLE_NAME, True],
        ["nonexistent-role-name", False],
    ],
)
def test_step_role_name_as_expected_filter(role_name_expected, success, sagemaker_pipeline) -> None:
    step_role_validation = StepRoleNameAsExpected(
        role_name_expected=role_name_expected,
        step_name="sm_processing_step_sklearn",
        rule=Equals(),
    )
    result = step_role_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "role_name_expected,success",
    [
        [TEST_ROLE_NAME, True],
        ["nonexistent-role-name", False],
    ],
)
def test_step_role_name_as_expected_no_filter(
    role_name_expected, success, sagemaker_pipeline
) -> None:
    step_role_validation = StepRoleNameAsExpected(
        role_name_expected=role_name_expected,
        rule=Equals(),
    )
    result = step_role_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "rule,tags_expected,step_name,success",
    [
        [Equals, [{"Key": "some-key", "Value": "some-value"}], "sm_processing_step_spark", True],
        [Contains, [{"Key": "some-key", "Value": "some-value"}], "sm_processing_step_spark", True],
        [
            Contains,
            [{"Key": "some-key", "Value": "some-value"}],
            "sm_processing_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                {
                    "Key": "some-key",
                    "Value": ParameterString(name="parameter-1", default_value="some-value-1"),
                }
            ],
            "sm_training_step_sklearn",
            False,
        ],
        [
            Equals,
            [
                {"Key": "some-key", "Value": "some-value"},
                {
                    "Key": "some-key",
                    "Value": ParameterString(name="parameter-1", default_value="some-value-1"),
                },
            ],
            "sm_training_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                {"Key": "some-key", "Value": "some-value"},
                {"Key": "some-non-existent-key", "Value": "some-non-existent-value"},
            ],
            "sm_processing_step_spark",
            False,
        ],
        [
            Equals,
            [{"Key": "some-key", "Value": "some-non-existent-value"}],
            "sm_processing_step_spark",
            False,
        ],
        [
            Equals,
            [{"Key": "some-non-existent-key", "Value": "some-value"}],
            "sm_processing_step_spark",
            False,
        ],
        [Equals, [], "sm_processing_step_spark", False],
    ],
)
def test_step_tags_as_expected_filter(
    rule, tags_expected, step_name, success, sagemaker_pipeline
) -> None:
    step_role_validation = StepTagsAsExpected(
        tags_expected=tags_expected,
        step_name=step_name,
        rule=rule(),
    )
    result = step_role_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "rule,tags_expected,success",
    [
        [Equals, [{"Key": "some-key", "Value": "some-value"}], False],
        [Equals, [{"Key": "some-non-existent-key", "Value": "some-non-existent-value"}], False],
        [
            Contains,
            [
                {
                    "Key": "some-key",
                    "Value": ParameterString(name="parameter-1", default_value="some-value-1"),
                }
            ],
            False,
        ],
        [Contains, [{"Key": "some-key", "Value": "some-value"}], True],
    ],
)
def test_step_tags_as_expected_no_filter(rule, tags_expected, success, sagemaker_pipeline) -> None:
    step_role_validation = StepTagsAsExpected(
        tags_expected=tags_expected,
        rule=rule(),
    )
    result = step_role_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "rule,inputs_expected,step_type,success",
    [
        [
            Contains,
            [
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/input-2",
                    destination="/opt/ml/processing/input",
                    input_name="input-2",
                )
            ],
            "Processing",
            True,
        ],
        [
            Contains,
            [
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/input-2",
                    destination="/opt/ml/processing/input",
                    input_name="input-3",
                )
            ],
            "Processing",
            False,
        ],
        [
            Contains,
            [
                {
                    "train": TrainingInput(
                        s3_data=f"s3://{DUMMY_BUCKET}/output-4",
                        content_type="text/csv",
                    )
                }
            ],
            "Training",
            True,
        ],
        [
            Contains,
            [
                {
                    "validation": FileSystemInput(
                        file_system_id="fs-1234",
                        file_system_type="EFS",
                        directory_path="/some/path",
                        file_system_access_mode="ro",
                    )
                }
            ],
            "Training",
            True,
        ],
    ],
)
def test_step_inputs_as_expected_no_filter(
    rule, inputs_expected, step_type, success, sagemaker_pipeline
) -> None:
    step_inputs_validation = StepInputsAsExpected(
        inputs_expected=inputs_expected,
        step_type=step_type,
        rule=rule(),
    )
    result = step_inputs_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "rule,inputs_expected,step_name,success",
    [
        [
            Equals,
            [
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/output-1",
                    destination="/opt/ml/processing/input",
                    input_name="input-1",
                ),
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/input-2",
                    destination="/opt/ml/processing/input",
                    input_name="input-2",
                ),
            ],
            "sm_processing_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/input-2",
                    destination="/opt/ml/processing/input",
                    input_name="input-2",
                ),
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/output-1",
                    destination="/opt/ml/processing/input",
                    input_name="input-1",
                ),
            ],
            "sm_processing_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                {
                    "train": TrainingInput(
                        s3_data=f"s3://{DUMMY_BUCKET}/output-4",
                        content_type="text/csv",
                    ),
                    "validation": FileSystemInput(
                        file_system_id="fs-1234",
                        file_system_type="EFS",
                        directory_path="/some/path",
                        file_system_access_mode="ro",
                    ),
                    "some-key": "some-value",
                }
            ],
            "sm_training_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                {
                    "train": TrainingInput(
                        s3_data=f"s3://{DUMMY_BUCKET}/output-4",
                        content_type="text/csv",
                    ),
                    "some-key": "some-value",
                    "validation": FileSystemInput(
                        file_system_id="fs-1234",
                        file_system_type="EFS",
                        directory_path="/some/path",
                        file_system_access_mode="ro",
                    ),
                }
            ],
            "sm_training_step_sklearn",
            True,
        ],
        [
            Equals,
            [
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/does-not-match",
                    destination="/opt/ml/processing/input",
                    input_name="input-1",
                ),
                ProcessingInput(
                    source=f"s3://{DUMMY_BUCKET}/input-1",
                    destination="/opt/ml/processing/input",
                    input_name="input-2",
                ),
            ],
            "sm_processing_step_sklearn",
            False,
        ],
    ],
)
def test_step_inputs_as_expected_filter(
    rule, inputs_expected, step_name, success, sagemaker_pipeline
) -> None:
    step_inputs_validation = StepInputsAsExpected(
        inputs_expected=inputs_expected,
        step_name=step_name,
        rule=rule(),
    )
    result = step_inputs_validation.run(sagemaker_pipeline)
    assert result.success == success


def test_step_inputs_as_expected_args_validation_step_type() -> None:
    with pytest.raises(ValueError):
        StepInputsAsExpected(
            inputs_expected=[],
            step_type="does-not-exist",
            rule=Equals(),
        )


def test_step_inputs_as_expected_args_validation_exclusive() -> None:
    with pytest.raises(ValueError):
        StepInputsAsExpected(
            inputs_expected=[],
            step_type="does-not-exist",
            step_name="does-not-exist",
            rule=Equals(),
        )


def test_step_inputs_as_expected_args_validation_neither() -> None:
    with pytest.raises(ValueError):
        StepInputsAsExpected(
            inputs_expected=[],
            rule=Equals(),
        )


@pytest.mark.parametrize(
    "rule,outputs_expected,success",
    [
        [
            Contains,
            [
                ProcessingOutput(
                    output_name="output-1",
                    source="/opt/ml/processing/output/1",
                    destination=f"s3://{DUMMY_BUCKET}/output-1",
                ),
                ProcessingOutput(
                    output_name="output-2",
                    source="/opt/ml/processing/output/2",
                    destination=f"s3://{DUMMY_BUCKET}/output-2",
                ),
            ],
            True,
        ],
        [
            Contains,
            [
                ProcessingOutput(
                    output_name="output-999",
                    source="/opt/ml/processing/output/1",
                    destination=f"s3://{DUMMY_BUCKET}/output-1",
                ),
                ProcessingOutput(
                    output_name="output-2",
                    source="/opt/ml/processing/output/2",
                    destination=f"s3://{DUMMY_BUCKET}/output-2",
                ),
            ],
            False,
        ],
        [
            Equals,
            [
                ProcessingOutput(
                    output_name="output-1",
                    source="/opt/ml/processing/output/1",
                    destination=f"s3://{DUMMY_BUCKET}/output-1",
                ),
                ProcessingOutput(
                    output_name="output-2",
                    source="/opt/ml/processing/output/2",
                    destination=f"s3://{DUMMY_BUCKET}/output-2",
                ),
            ],
            False,
        ],
    ],
)
def test_step_outputs_as_expected_no_filter(
    rule, outputs_expected, success, sagemaker_pipeline
) -> None:
    step_outputs_validation = StepOutputsAsExpected(
        outputs_expected=outputs_expected,
        rule=rule(),
    )
    result = step_outputs_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "rule,outputs_expected,step_name,success",
    [
        [
            Equals,
            [
                ProcessingOutput(
                    output_name="output-1",
                    source="/opt/ml/processing/output/1",
                    destination=f"s3://{DUMMY_BUCKET}/output-1",
                ),
                ProcessingOutput(
                    output_name="output-2",
                    source="/opt/ml/processing/output/2",
                    destination=f"s3://{DUMMY_BUCKET}/output-2",
                ),
                ProcessingOutput(
                    output_name="output-3",
                    source="/opt/ml/processing/output/3",
                    destination=Join(
                        on="/",
                        values=[
                            "s3:/",
                            DUMMY_BUCKET,
                            ExecutionVariables.PIPELINE_EXECUTION_ID,
                            "output-3",
                        ],
                    ),
                ),
            ],
            "sm_processing_step_sklearn",
            True,
        ],
        [
            Contains,
            [
                ProcessingOutput(
                    output_name="output-2",
                    source="/opt/ml/processing/output/2",
                    destination=f"s3://{DUMMY_BUCKET}/output-2",
                ),
            ],
            "sm_processing_step_sklearn",
            True,
        ],
    ],
)
def test_step_outputs_as_expected_filter(
    rule, outputs_expected, step_name, success, sagemaker_pipeline
) -> None:
    step_outputs_validation = StepOutputsAsExpected(
        outputs_expected=outputs_expected,
        step_name=step_name,
        rule=rule(),
    )
    result = step_outputs_validation.run(sagemaker_pipeline)
    assert result.success == success


@pytest.mark.parametrize(
    "inputs_outputs_expected,success,raise_error",
    [
        [
            [
                {
                    "input": {
                        "step_name": "sm_processing_step_sklearn",
                        "input_name": "input-1",
                    },
                    "output": {
                        "step_name": "sm_processing_step_sklearn",
                        "output_name": "output-1",
                    },
                }
            ],
            True,
            False,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_processing_step_sklearn",
                        "input_name": "input-1",
                    },
                    "output": {
                        "step_name": "sm_processing_step_sklearn",
                        "output_name": "output-2",
                    },
                }
            ],
            False,
            False,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_processing_step_spark",
                        "input_name": "input-3",
                    },
                    "output": {
                        "step_name": "sm_processing_step_sklearn",
                        "output_name": "output-3",
                    },
                }
            ],
            True,
            False,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_processing_step_sklearn",
                        "input_name": "input-1",
                    },
                    "output": {
                        "step_name": "step_doesnt_exist",
                        "output_name": "output-1",
                    },
                }
            ],
            False,
            True,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_processing_step_sklearn",
                        "input_name": "input-doesnt-exist1",
                    },
                    "output": {
                        "step_name": "sm_processing_step_sklearn",
                        "output_name": "output-1",
                    },
                }
            ],
            False,
            True,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_training_step_sklearn",
                        "input_name": "train",
                    },
                    "output": {
                        "step_name": "sm_processing_step_spark",
                        "output_name": "output-2",
                    },
                }
            ],
            True,
            False,
        ],
        [
            [
                {
                    "input": {
                        "step_name": "sm_training_step_sklearn",
                        "input_name": "validation",
                    },
                    "output": {
                        "step_name": "sm_processing_step_spark",
                        "output_name": "output-2",
                    },
                }
            ],
            False,
            True,
        ],
    ],
)
def test_step_outputs_match_inputs_as_expected(
    sagemaker_pipeline, inputs_outputs_expected, success, raise_error
) -> None:
    step_outputs_validation = StepOutputsMatchInputsAsExpected(
        inputs_outputs_expected=inputs_outputs_expected,
    )
    if raise_error:
        with pytest.raises((ValueError, AttributeError)):
            _ = step_outputs_validation.run(sagemaker_pipeline)
    else:
        result = step_outputs_validation.run(sagemaker_pipeline)
        assert result.success is success


@mock_sqs
@mock_iam
def test_sqs_queue_exists_positive(sqs_client, iam_client, sagemaker_pipeline) -> None:
    with create_sqs_queue(sqs_client, iam_client, [(TEST_SQS_QUEUE_NAME, TEST_SQS_QUEUE_URL_BASE)]):
        sqs_queue_exists = StepCallbackSqsQueueExists()
        result = sqs_queue_exists.run(sagemaker_pipeline)
    assert result.success


@mock_sqs
def test_sqs_queue_exists_negative(sqs_client, iam_client, sagemaker_pipeline) -> None:
    sqs_queue_exists = StepCallbackSqsQueueExists()
    result = sqs_queue_exists.run(sagemaker_pipeline)
    assert not result.success



import pandas as pd
import pytest
from moto import mock_ecr

from sagemaker_rightline.model import (
    Configuration,
    Report,
    Validation,
    ValidationFailedError,
)
from sagemaker_rightline.rules import Equals, Rule
from sagemaker_rightline.validations import (
    ContainerImage,
    StepImagesExist,
    StepKmsKeyIdAsExpected,
    ValidationResult,
)
from tests.fixtures.image_details import IMAGE_1_URI, IMAGE_2_URI
from tests.fixtures.pipeline import get_sagemaker_pipeline
from tests.utils import create_image, ecr_client


@pytest.fixture
def sagemaker_pipeline():
    return get_sagemaker_pipeline()


@pytest.mark.parametrize(
    "validations,error",
    [
        [["not_validation"], True],
        ["not_list", True],
        [[], True],
        [[StepImagesExist()], False],
        [[StepImagesExist(), StepImagesExist()], False],
    ],
)
def test_configuration_validate_input_validations(validations, error: bool) -> None:
    if error:
        with pytest.raises((ValueError, TypeError)):
            Configuration._validate_input_validations(validations=validations)
    else:
        Configuration._validate_input_validations(validations=validations)


@pytest.mark.parametrize(
    "results",
    [
        [
            ValidationResult(
                success=True,
                negative=False,
                message="test-message",
                subject="test-subject-1",
                validation_name="test",
            )
        ],
        [
            ValidationResult(
                success=True,
                negative=False,
                message="test-message-true",
                subject="test-subject-2",
                validation_name="test",
            ),
            ValidationResult(
                success=False,
                negative=False,
                message="test-message-false",
                subject="test-subject-3",
                validation_name="test",
            ),
        ],
    ],
)
def test_configuration_make_report(results) -> None:
    report = Configuration._make_report(results=results)
    assert report.results == results
    assert isinstance(report, Report)
    assert isinstance(report.results, list)
    assert isinstance(report.results[0], ValidationResult)


@mock_ecr
@pytest.mark.parametrize(
    "expected_report_length,validations, return_df",
    [
        [1, [StepImagesExist()], False],
        [2, [StepImagesExist(), StepImagesExist()], True],
    ],
)
def test_configuration_run(
    sagemaker_pipeline, ecr_client, expected_report_length, validations, return_df
) -> None:
    """Test run method of Configuration class."""
    container_images = [
        ContainerImage(uri=IMAGE_1_URI),
        ContainerImage(uri=IMAGE_2_URI),
    ]
    with create_image(ecr_client, container_images):
        cf = Configuration(validations=validations, sagemaker_pipeline=sagemaker_pipeline)
        report = cf.run(fail_fast=False, return_df=return_df)

    if return_df:
        assert isinstance(report, pd.DataFrame)
        observed_report_len = len(report)
    else:
        assert isinstance(report, Report)
        observed_report_len = len(report.results)
    assert observed_report_len == expected_report_length


def test_configuration_handle_empty_results(sagemaker_pipeline) -> None:
    """Test run method of Configuration class."""
    validation = StepKmsKeyIdAsExpected(
        step_name="sm_processing_step_sklearn",
        rule=Equals(),
        kms_key_id_expected="some/kms-key-alias",
    )
    cf = Configuration(validations=[validation], sagemaker_pipeline=sagemaker_pipeline)
    validation_result = cf._handle_empty_results(
        result=None,
        validation=validation,
    )
    assert isinstance(validation_result, ValidationResult)
    assert "not return any results" in validation_result.message
    assert not validation_result.success


@mock_ecr
@pytest.mark.parametrize(
    "raises_error,validations",
    [
        [False, [StepImagesExist()]],
        [True, [StepImagesExist(), StepImagesExist()]],
    ],
)
def test_configuration_run_fail_fast(sagemaker_pipeline, raises_error, validations) -> None:
    """Test run method of Configuration class."""
    cf = Configuration(validations=validations, sagemaker_pipeline=sagemaker_pipeline)
    if raises_error:
        with pytest.raises(ValidationFailedError):
            cf.run(fail_fast=True)
    else:
        report = cf.run(fail_fast=True)
        assert len(report.results) == 1


def test_report_to_df() -> None:
    """Test to_df method of Report class."""
    validation_name = "StepImagesExist"
    report = Report(
        results=[
            ValidationResult(
                success=True,
                negative=False,
                message="test-message-0",
                subject="test-subject-0",
                validation_name=validation_name,
            ),
        ]
    )
    df = report.to_df()
    assert df.shape == (1, 5)
    assert df.columns.tolist() == list(ValidationResult.__annotations__.keys())
    assert df["success"].tolist() == [True]
    assert df["negative"].tolist() == [False]
    assert df["message"].tolist() == ["test-message-0"]
    assert df["validation_name"].tolist() == [validation_name]


def test_validation_get_attribute_filter(sagemaker_pipeline) -> None:
    """Test get_attribute method of Validation class."""
    kms_key_alias_expected = "some/kms-key-alias"
    validation = StepKmsKeyIdAsExpected(
        step_name="sm_processing_step_sklearn",
        rule=Equals(),
        kms_key_id_expected=kms_key_alias_expected,
    )
    assert Validation.get_attribute(sagemaker_pipeline, validation.paths) == [
        kms_key_alias_expected
    ]


def test_validation_get_attribute_no_filter(sagemaker_pipeline) -> None:
    """Test get_attribute method of Validation class."""
    kms_key_alias_expected = "some/kms-key-alias"
    validation = StepKmsKeyIdAsExpected(
        rule=Equals(),
        kms_key_id_expected=kms_key_alias_expected,
    )
    assert Validation.get_attribute(sagemaker_pipeline, validation.paths) == [
        kms_key_alias_expected,
        kms_key_alias_expected,
        kms_key_alias_expected,
    ]

    validation = StepImagesExist()
    assert Validation.get_attribute(sagemaker_pipeline, validation.paths) == [
        IMAGE_1_URI,
        IMAGE_2_URI,
        IMAGE_1_URI,
    ]


def test_validation_failed_error():
    """Test ValidationFailedError class."""
    validation_result = ValidationResult(
        success=False,
        negative=False,
        message="test-message",
        subject="test-subject",
        validation_name="test",
    )
    with pytest.raises(ValidationFailedError):
        try:
            raise ValidationFailedError(validation_result)
        except ValidationFailedError as e:
            assert isinstance(e.message, str)
            assert e.validation_result == validation_result
            assert isinstance(e, ValidationFailedError)
            assert isinstance(e, Exception)
            raise



import re
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import boto3
from botocore.exceptions import ClientError
from sagemaker.estimator import Estimator
from sagemaker.inputs import FileSystemInput, TrainingInput
from sagemaker.processing import NetworkConfig, ProcessingInput, ProcessingOutput
from sagemaker.workflow.entities import PipelineVariable
from sagemaker.workflow.parameters import Parameter
from sagemaker.workflow.pipeline import Pipeline

from sagemaker_rightline.model import Rule, Validation, ValidationResult
from sagemaker_rightline.rules import Equals


class PipelineParametersAsExpected(Validation):
    """Validate Pipeline Parameters.

    This validation will check if the parameters in the pipeline are as
    expected.
    """

    def __init__(
        self, parameters_expected: List[Parameter], rule: Rule, ignore_default_value: bool = False
    ) -> None:
        """Validate Pipeline Parameters."""
        super().__init__(name="PipelineParametersAsExpected", paths=[".parameters[]"], rule=rule)
        if not parameters_expected:
            raise ValueError("parameters_expected cannot be empty.")
        self.parameters_expected: List[Parameter] = parameters_expected
        self.ignore_default_value: bool = ignore_default_value

    def run(self, sagemaker_pipeline: Pipeline) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: Dict containing ValidationResult
        :rtype: ValidationResult
        """
        parameters_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        if self.ignore_default_value:
            for ix, parameter in enumerate(parameters_observed):
                parameters_observed[ix].default_value = None
        result = self.rule.run(parameters_observed, self.parameters_expected, self.name)
        return result


class StepKmsKeyIdAsExpected(Validation):
    """Validate KmsKeyId or output_kms_key in Step.

    Supported only for ProcessingStep and TrainingStep (output_kms_key).
    This validation is useful when you want to ensure that the KmsKeyId
    or output_kms_key of a Pipeline Step is as expected.
    """

    def __init__(
        self, kms_key_id_expected: str, rule: Rule, step_name: Optional[str] = None
    ) -> None:
        """Initialize StepKmsKeyIdAsExpected validation."""
        self.step_filter: str = f"name=={step_name}" if step_name else ""
        super().__init__(
            name="StepKmsKeyIdAsExpected",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing].kms_key",
                f".steps[{self.step_filter} && step_type/value==Training].estimator.output_kms_key",
            ],
            rule=rule,
        )
        self.kms_key_id_expected: str = kms_key_id_expected

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        kms_keys_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        result = self.rule.run(kms_keys_observed, [self.kms_key_id_expected], self.name)
        return result


@dataclass
class ContainerImage:
    """Container Image dataclass."""

    uri: str

    def __post_init__(self) -> None:
        """Decompose ImageUri into its components."""
        self.account_id = self.uri.split(".")[0]
        self.repository = re.search("/(.*):", self.uri).group(1)
        self.region = self.uri.split(".")[3]
        self.tag = self.uri.split(":")[1]


class StepImagesExist(Validation):
    """Check if container images exist in ECR.

    Supported only for ProcessingStep and TrainingStep.
    """

    def __init__(
        self,
        boto3_client: Union["boto3.client('ecr')", str] = "ecr",  # noqa F821
        step_name: Optional[str] = None,
    ) -> None:
        """Initialize ImageExists validation.

        :param boto3_client: boto3 client to use, defaults to "ecr"
        :type boto3_client: Union["boto3.client('ecr')", str], optional
        :param step_name: Step name to filter on, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """
        self.step_filter: str = f"name=={step_name}" if step_name else ""
        super().__init__(
            name="StepImagesExist",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing].processor.image_uri",
                f".steps[{self.step_filter} && step_type/value==Training].estimator.image_uri",
            ],
        )
        if isinstance(boto3_client, str) and boto3_client != "ecr":
            raise ValueError(f"boto3_client must be 'ecr', not {boto3_client}.")
        if boto3_client:
            self.client = (
                boto3_client if not isinstance(boto3_client, str) else boto3.client(boto3_client)
            )

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of whether Image URIs referenced in Pipeline Steps
        exist in ECR.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        paginator = self.client.get_paginator("describe_images")

        uris = Validation.get_attribute(sagemaker_pipeline, self.paths)
        # return uris
        not_exist = []
        exist = []
        for uri in uris:
            container_image = ContainerImage(uri=uri)
            try:
                _ = list(
                    paginator.paginate(
                        repositoryName=container_image.repository,
                        registryId=container_image.account_id,
                        imageIds=[{"imageTag": container_image.tag}],
                    )
                )
                exist.append(uri)
            except ClientError:
                not_exist.append(uri)
        if not_exist:
            return ValidationResult(
                validation_name=self.name,
                success=False,
                negative=False,
                message=f"Images {not_exist} do not exist.",
                subject=str(not_exist),
            )
        return ValidationResult(
            validation_name=self.name,
            success=True,
            negative=False,
            message=f"Images {exist} exist.",
            subject=str(exist),
        )


class StepNetworkConfigAsExpected(Validation):
    """Validate NetworkConfig in Step.

    This Validation currently supports only ProcessingStep. This
    validation is useful when you want to ensure that the NetworkConfig
    of a Pipeline Step's Processor is as expected.
    """

    def __init__(
        self, network_config_expected: NetworkConfig, rule: Rule, step_name: Optional[str] = None
    ) -> None:
        """Initialize StepNetworkConfigAsExpected validation.

        :param network_config_expected: Expected NetworkConfig
        :type network_config_expected: NetworkConfig
        :param rule: Rule to apply
        :type rule: Rule
        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """
        self.step_filter: str = f"name=={step_name}" if step_name else ""
        super().__init__(
            name="StepNetworkConfigAsExpected",
            paths=[
                f".steps[{self.step_filter}].processor.network_config",
            ],
            rule=rule,
        )
        self.network_config_expected: NetworkConfig = network_config_expected

    @staticmethod
    def get_training_step_network_config(
        training_step_estimators: List[Estimator],
    ) -> List[NetworkConfig]:
        """Get NetworkConfig of each training step estimator.

        :param training_step_estimators: List of training step estimators
        :type training_step_estimators: List[Estimator]
        :return: List of NetworkConfig of each training step estimator
        :rtype: List[NetworkConfig]
        """
        default_network_config = NetworkConfig()
        training_step_network_configs = []
        for step in training_step_estimators:
            step_dict = {}
            for attr_name in default_network_config.__dict__.keys():
                attr_value = getattr(step, attr_name)
                # Some attributes of NetworkConfig are callable and return the value,
                # so we need to call them
                step_dict[attr_name] = (
                    attr_value
                    if any([isinstance(attr_value, bool), isinstance(attr_value, list)])
                    else attr_value()
                )
            training_step_network_configs.append(NetworkConfig(**step_dict))
        return training_step_network_configs

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of NetworkConfigs on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        network_configs_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)

        # Compatibility with TrainingStep, which does not have a NetworkConfig object
        # as attribute, but takes the attributes of NetworkConfig as individual arguments.
        training_step_estimators = [
            step.estimator
            for step in sagemaker_pipeline.steps
            if step.step_type.value == "Training"
            and step.name == self.step_filter.replace("name==", "")
        ]
        if training_step_estimators:
            network_configs_observed += (
                StepNetworkConfigAsExpected.get_training_step_network_config(
                    training_step_estimators
                )
            )

        network_configs_observed_dict = []
        for nwc in network_configs_observed:
            if nwc:
                network_configs_observed_dict.append(nwc.__dict__)
            else:
                network_configs_observed_dict.append(None)
        network_config_expected_dict = (
            self.network_config_expected.__dict__
            if self.network_config_expected
            else self.network_config_expected
        )
        result = self.rule.run(
            network_configs_observed_dict, [network_config_expected_dict], self.name
        )
        return result


class StepLambdaFunctionExists(Validation):
    """Validate whether Lambda Function referenced in LambdaSteps exists."""

    def __init__(
        self,
        boto3_client: Union["boto3.client('lambda')", str] = "lambda",  # noqa F821
    ) -> None:
        """Initialize StepLambdaFunctionArnExists validation.

        :param boto3_client: Boto3 client to use for checking Lambda Function existence,
        defaults to "lambda"
        :type boto3_client: Union["boto3.client('lambda')", str], optional
        :raises ValueError: boto3_client must be 'lambda'
        :return: None
        :rtype: None
        """
        if isinstance(boto3_client, str) and boto3_client != "lambda":
            raise ValueError(f"boto3_client must be 'lambda', not {boto3_client}.")
        if boto3_client:
            self.client = (
                boto3_client if not isinstance(boto3_client, str) else boto3.client(boto3_client)
            )

        super().__init__(
            name="StepLambdaFunctionExists",
            paths=[
                ".steps[step_type/value==Lambda].lambda_func",
            ],
        )

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        lambda_func_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        exist = []
        not_exist = []
        for func in lambda_func_observed:
            try:
                _ = self.client.get_function(
                    FunctionName=func,
                )
                exist.append(func)
            except ClientError:
                not_exist.append(func)
        if not_exist:
            return ValidationResult(
                success=False,
                negative=False,
                message=f"Lambda Function {not_exist} does not " f"exist.",
                subject=str(lambda_func_observed),
                validation_name=self.name,
            )
        else:
            return ValidationResult(
                success=True,
                negative=False,
                message=f"Lambda Function {exist} exists.",
                subject=str(lambda_func_observed),
                validation_name=self.name,
            )


class StepRoleNameAsExpected(Validation):
    """Validate Role of Pipeline Step.

    Supported only for ProcessingStep and TrainingStep. This validation
    is useful when you want to ensure that the Role of a Pipeline Step
    is as expected.
    """

    def __init__(
        self,
        role_name_expected: str,
        rule: Rule,
        step_name: Optional[str] = None,
    ) -> None:
        """Initialize StepRole validation.

        :param role_name_expected: Expected Role name
        :type role_name_expected: str
        :param rule: Rule to use for validation
        :type rule: Rule
        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """
        self.step_filter: str = f"name=={step_name}" if step_name else ""

        super().__init__(
            name="StepRole",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing].processor.role",
                f".steps[{self.step_filter} && step_type/value==Training].estimator.role",
            ],
            rule=rule,
        )
        self.role_name_expected: str = role_name_expected

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        role_arns_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        role_name_observed = [role_arn.split("/")[-1] for role_arn in role_arns_observed]
        result = self.rule.run(role_name_observed, [self.role_name_expected], self.name)
        return result


class StepRoleNameExists(Validation):
    """Validate existence of Role of Pipeline Step.

    Supported only for ProcessingStep and TrainingStep. This validation
    is useful when you want to ensure that Roles of Pipeline Steps
    exist.
    """

    def __init__(
        self,
        step_name: Optional[str] = None,
        boto3_client: Union["boto3.client('iam')", str] = "iam",  # noqa F821
    ) -> None:
        """Initialize StepRole validation.

        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :param boto3_client: Boto3 client to use for checking Role existence, defaults to "iam"
        :type boto3_client: Union["boto3.client('iam')", str], optional
        :return: None
        :rtype: None
        """
        self.step_filter: str = f"name=={step_name}" if step_name else ""
        if isinstance(boto3_client, str) and boto3_client != "iam":
            raise ValueError(f"boto3_client must be 'iam', not {boto3_client}.")
        if boto3_client:
            self.client = (
                boto3_client if not isinstance(boto3_client, str) else boto3.client(boto3_client)
            )

        super().__init__(
            name="StepRoleExists",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing].processor.role",
                f".steps[{self.step_filter} && step_type/value==Training].estimator.role",
            ],
        )

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        role_arns_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        role_name_observed = [role_arn.split("/")[-1] for role_arn in role_arns_observed]

        exist = []
        not_exist = []
        for role_name in role_name_observed:
            try:
                _ = self.client.get_role(
                    RoleName=role_name,
                )
                exist.append(role_name)
            except ClientError:
                not_exist.append(role_name)
        if not_exist:
            return ValidationResult(
                success=False,
                negative=False,
                message=f"Role {not_exist} does not exist.",
                subject=str(role_name_observed),
                validation_name=self.name,
            )
        else:
            return ValidationResult(
                success=True,
                negative=False,
                message=f"Role {exist} exists.",
                subject=str(role_name_observed),
                validation_name=self.name,
            )


class StepTagsAsExpected(Validation):
    """Validate Tags of Pipeline Step.

    Supported only for ProcessingStep and TrainingStep. This validation
    is useful when you want to ensure that the Tags of a Pipeline Step
    are as expected.
    """

    def __init__(
        self,
        tags_expected: List[Dict[str, Union[str, PipelineVariable]]],
        rule: Rule,
        step_name: Optional[str] = None,
    ) -> None:
        """Initialize StepTagsAsExpected validation.

        :param tags_expected: Expected Tags
        :type tags_expected: List[Dict[str, Union[str, PipelineVariable]]]
        :param rule: Rule to use for validation
        :type rule: Rule
        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """
        self.step_filter: str = f"name=={step_name}" if step_name else ""

        super().__init__(
            name="StepTagsAsExpected",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing].processor.tags",
                f".steps[{self.step_filter} && step_type/value==Training].estimator.tags",
            ],
            rule=rule,
        )
        self.tags_expected: List[Dict[str, Union[str, PipelineVariable]]] = tags_expected

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Parameters on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        tags_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        result = self.rule.run(tags_observed[0], self.tags_expected, self.name)
        return result


class StepInputsAsExpected(Validation):
    """Validate Inputs of Pipeline Step.

    Supported only for ProcessingStep and TrainingStep. This validation
    is useful when you want to ensure that the Inputs of a Pipeline Step
    are as expected.
    """

    def __init__(
        self,
        inputs_expected: List[
            Union[Dict[str, Union[str, TrainingInput, FileSystemInput]], ProcessingInput]
        ],
        rule: Rule,
        step_type: Optional[Union["Training", "Processing"]] = None,  # noqa F821
        step_name: Optional[str] = None,
    ) -> None:
        """Initialize StepTagsAsExpected validation.

        :param inputs_expected: Expected Inputs
        :type inputs_expected: List[Dict[str, Union[str, TrainingInput, ProcessingInput,
        FileSystemInput]]]
        :param rule: Rule to use for validation
        :type rule: Rule
        :param step_type: Type of Step to validate
        :type step_type: Union["Training", "Processing"]
        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """
        if step_type and step_name:
            raise ValueError("step_type and step_name cannot be specified together.")
        if not step_type and not step_name:
            raise ValueError("Either step_type or step_name must be specified.")
        if step_type not in ("Training", "Processing") and not step_name:
            raise ValueError(f"step_type must be 'Training' or 'Processing', not {step_type}.")

        self.step_filter: str = f"name=={step_name}" if step_name else ""
        self.step_type_filter: str = f"step_type/value=={step_type}" if step_type else ""

        super().__init__(
            name="StepInputsAsExpected",
            paths=[
                f".steps[{self.step_filter} && {self.step_type_filter}].inputs",
            ],
            rule=rule,
        )
        self.inputs_expected: List[
            Dict[str, Union[str, TrainingInput, ProcessingInput, FileSystemInput]]
        ] = inputs_expected

    @staticmethod
    def format_training_inputs(
        inputs: List[Dict[str, Union[str, TrainingInput, FileSystemInput]]]
    ) -> List[Dict[str, Union[dict, str]]]:
        """Format Training Inputs.

        :param inputs: Training Inputs
        :type inputs: List[Dict[str, Union[str, TrainingInput, FileSystemInput]]]
        :return: Formatted Training Inputs
        :rtype: List[Dict[str, Union[dict, str]]]
        """
        formatted_inputs = []
        for item in inputs:
            for key, value in item.items():
                formatted_inputs.append(
                    {
                        key: value.__dict__
                        if isinstance(value, TrainingInput) or isinstance(value, FileSystemInput)
                        else value
                    }
                )
        return formatted_inputs

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Step Inputs on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        if self.step_filter:
            step_type = Validation.get_attribute(
                sagemaker_pipeline, [f".steps[{self.step_filter}].step_type.value"]
            )[0]
            self.step_type_filter = f"step_type/value=={step_type}"

        inputs_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)

        if self.step_type_filter == "step_type/value==Processing":
            # ProcessingStep has a list of ProcessingInput
            inputs_observed_formatted = [x.__dict__ for y in inputs_observed for x in y]
            inputs_expected_formatted = [x.__dict__ for x in self.inputs_expected]
        elif self.step_type_filter == "step_type/value==Training":
            # TrainingStep has a dict with values potentially being TrainingInput or FileSystemInput
            inputs_observed_formatted = StepInputsAsExpected.format_training_inputs(inputs_observed)
            inputs_expected_formatted = StepInputsAsExpected.format_training_inputs(
                self.inputs_expected
            )
        result = self.rule.run(inputs_observed_formatted, inputs_expected_formatted, self.name)
        return result


class StepOutputsAsExpected(Validation):
    """Validate Outputs of Pipeline Step.

    Supported only for ProcessingStep. This validation is useful when
    you want to ensure that the Outputs of a Pipeline Step are as
    expected.
    """

    def __init__(
        self,
        outputs_expected: List[ProcessingOutput],
        rule: Rule,
        step_name: Optional[str] = None,
    ) -> None:
        """Initialize StepTagsAsExpected validation.

        :param outputs_expected: Expected Inputs
        :type outputs_expected: List[ProcessingOutput]
        :param rule: Rule to use for validation
        :type rule: Rule
        :param step_name: Name of Step to validate, defaults to None
        :type step_name: Optional[str], optional
        :return: None
        :rtype: None
        """

        self.step_filter: str = f"name=={step_name}" if step_name else ""

        super().__init__(
            name="StepOutputsAsExpected",
            paths=[
                f".steps[{self.step_filter} && step_type/value==Processing]].outputs",
            ],
            rule=rule,
        )
        self.outputs_expected: List[ProcessingOutput] = outputs_expected

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of StepOutputs on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """

        outputs_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        outputs_observed_formatted = [x.__dict__ for y in outputs_observed for x in y]
        outputs_expected_formatted = [x.__dict__ for x in self.outputs_expected]
        result = self.rule.run(outputs_observed_formatted, outputs_expected_formatted, self.name)
        return result


class StepOutputsMatchInputsAsExpected(Validation):
    """Validate Outputs <> Inputs of Pipeline Step."""

    def __init__(self, inputs_outputs_expected: List[Dict[str, Dict[str, str]]]) -> None:
        """Initialize StepTagsAsExpected validation.

        :param inputs_outputs_expected: Expected Input and Output per step name
        :type inputs_outputs_expected: Dict[str, Dict[str, str]]
        :return: None
        :rtype: None
        """
        name = "StepOutputsMatchInputsAsExpected"
        super().__init__(
            name=name,
            rule=Equals(),
        )
        self.inputs_outputs_expected: List[Dict[str, Dict[str, str]]] = inputs_outputs_expected

    @staticmethod
    def get_step_by_name(
        sagemaker_pipeline: Pipeline, step_name: str
    ) -> "ProcessingStep":  # noqa F821
        """Get ProcessingStep, TrainingStep by name.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :param step_name: Name of Step
        :type step_name: str
        :return: ProcessingStep
        :rtype: ProcessingStep
        """
        supported_step_types = ("Processing", "Training")
        for step in sagemaker_pipeline.steps:
            if step.name == step_name and step.step_type.value in supported_step_types:
                return step
        raise ValueError(f"Processing or Training Step {step_name} not found in Pipeline.")

    @staticmethod
    def get_input_output_by_name(
        step: Union["ProcessingStep", "TrainingStep"], name: str, kind: str  # noqa F821
    ) -> str:
        """Get ProcessingInput or ProcessingOutput by name.

        :param step: ProcessingStep
        :type step: ProcessingStep
        :param name: Name of Input or Output
        :type name: str
        :param kind: Kind of Input or Output, must be one of "input" or "output"
        :type kind: str
        :return: Input or Output path
        :rtype: str
        """
        step_type = step.step_type.value

        if kind == "input":
            if step_type == "Training":
                input = step.inputs[name]
                if isinstance(input, TrainingInput):
                    return input.config["DataSource"]["S3DataSource"]["S3Uri"]
                else:
                    raise ValueError(f"Input {name} is not of type TrainingInput.")
            else:
                # If ProcessingStep
                for input in step.inputs:
                    if input.input_name == name:
                        return input.source
                raise ValueError(f"Input {name} not found in ProcessingStep.")

        elif kind == "output":
            for output in step.outputs:
                if output.output_name == name:
                    return output.destination
        else:
            raise ValueError(f"Kind {kind} not supported. Must be one of 'input' or 'output'.")

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation of Step Inputs on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """

        input_kw = "input"
        output_kw = "output"

        results = []
        for pair in self.inputs_outputs_expected:
            input_step_name = pair[input_kw]["step_name"]
            input_name = pair[input_kw]["input_name"]
            output_step_name = pair[output_kw]["step_name"]
            output_name = pair[output_kw]["output_name"]

            input_step = StepOutputsMatchInputsAsExpected.get_step_by_name(
                sagemaker_pipeline, input_step_name
            )
            output_step = StepOutputsMatchInputsAsExpected.get_step_by_name(
                sagemaker_pipeline, output_step_name
            )

            input = StepOutputsMatchInputsAsExpected.get_input_output_by_name(
                input_step, input_name, input_kw
            )
            output = StepOutputsMatchInputsAsExpected.get_input_output_by_name(
                output_step, output_name, output_kw
            )
            results.append((input, output))
        return self.rule.run(
            observed=[x[0] for x in results],
            expected=[x[1] for x in results],
            validation_name=self.name,
        )


class StepCallbackSqsQueueExists(Validation):
    """Validate whether the SQS queue targeted by a CallbackStep exists."""

    def __init__(
        self,
        boto3_client: Union["boto3.client('sqs')", str] = "sqs",  # noqa F821
    ) -> None:
        """Initialize StepCallbackSqsQueueExists validation.

        :param boto3_client: Boto3 client to use for checking SQS queue existence,
        defaults to "sqs"
        :type boto3_client: Union["boto3.client('sqs')", str], optional
        :raises ValueError: boto3_client must be 'sqs'
        :return: None
        :rtype: None
        """
        if isinstance(boto3_client, str) and boto3_client != "sqs":
            raise ValueError(f"boto3_client must be 'sqs', not {boto3_client}.")
        if boto3_client:
            self.client = (
                boto3_client if not isinstance(boto3_client, str) else boto3.client(boto3_client)
            )

        super().__init__(
            name="StepCallbackSqsQueueExists",
            paths=[
                ".steps[step_type/value==Callback].sqs_queue_url",
            ],
        )

    def run(
        self,
        sagemaker_pipeline: Pipeline,
    ) -> ValidationResult:
        """Runs validation on Pipeline.

        :param sagemaker_pipeline: SageMaker Pipeline
        :type sagemaker_pipeline: sagemaker.workflow.pipeline.Pipeline
        :return: validation result
        :rtype: ValidationResult
        """
        sqs_url_observed = Validation.get_attribute(sagemaker_pipeline, self.paths)
        exist = []
        not_exist = []
        for url in sqs_url_observed:
            try:
                _ = self.client.get_queue_attributes(
                    QueueUrl=url,
                )
                exist.append(url)
            except ClientError:
                not_exist.append(url)
        if not_exist:
            return ValidationResult(
                success=False,
                negative=False,
                message=f"SQS url {not_exist} does not exist.",
                subject=str(sqs_url_observed),
                validation_name=self.name,
            )
        else:
            return ValidationResult(
                success=True,
                negative=False,
                message=f"SQS url Function {exist} exists.",
                subject=str(sqs_url_observed),
                validation_name=self.name,
            )



from typing import Any, List, Union

from sagemaker_rightline.model import Rule, ValidationResult


class Equals(Rule):
    """Check if two lists are equal."""

    def __init__(self, negative: bool = False) -> None:
        """Check if two lists are equal.

        :param negative: whether the rule should be inverted, i.e. "not" (default: False)
        :type negative: bool
        :return:
        :rtype:
        """
        super().__init__("Equals", negative)

    def run(
        self,
        observed: List[Union[int, float, str, dict]],
        expected: List[Union[int, float, str, dict]],
        validation_name: str,
    ) -> ValidationResult:
        """Check if two lists are equal.

        :param observed: observed list
        :type observed: List[Any]
        :param expected: expected list
        :type expected: List[Any]
        :param validation_name: name of the validation
        :type validation_name: str
        :return: validation result
        :rtype: ValidationResult
        """
        try:
            # In case of int, float, str
            is_equal = set(observed) == set(expected)
        except TypeError:
            # In case of dict
            if isinstance(observed, dict) and isinstance(expected, dict):
                is_equal = observed == expected
            # In case of nested list
            elif isinstance(observed, list) and isinstance(expected, list):
                is_equal = all(True if item in observed else False for item in expected) and all(
                    True if item in expected else False for item in observed
                )
            else:
                is_equal = observed == expected

        is_equal = is_equal if not self.negative else not is_equal
        return ValidationResult(
            validation_name=validation_name,
            success=is_equal,
            negative=self.negative,
            message=f"{str(observed)} does {'not ' if not is_equal else ''}equal {str(expected)}",
            subject=str(expected),
        )


class Contains(Rule):
    """Check if a list contains another list."""

    def __init__(self, negative: bool = False) -> None:
        """Check if a list contains another list.

        :param negative: whether the rule should be inverted, i.e. "not" (default: False)
        :type negative: bool
        :return:
        :rtype:
        """
        super().__init__("Contains", negative)

    def run(
        self, observed: List[Any], expected: List[Any], validation_name: str
    ) -> ValidationResult:
        """Check if a list contains another list.

        :param observed: observed list
        :type observed: List[Any]
        :param expected: expected list
        :type expected: List[Any]
        :param validation_name: name of the validation
        :type validation_name: str
        :return: validation result
        :rtype: ValidationResult
        """
        is_contained = all(item in observed for item in expected)
        is_contained = is_contained if not self.negative else not is_contained
        return ValidationResult(
            validation_name=validation_name,
            success=is_contained,
            negative=self.negative,
            message=f"{str(observed)} does {'not ' if not is_contained else ''}contain "
            f"{str(expected)}",
            subject=str(expected),
        )



from typing import List

import pytest
from sagemaker.workflow.parameters import ParameterString

from sagemaker_rightline.rules import Contains, Equals


@pytest.mark.parametrize(
    "observed,expected,negative,success",
    [
        [["test1"], ["test1"], False, True],
        [["test1"], ["test1"], True, False],
        [["test1"], ["test2"], False, False],
        [["test1"], ["test2"], True, True],
        [
            [ParameterString(name="test1"), ParameterString(name="test2")],
            [ParameterString(name="test1"), ParameterString(name="test2")],
            False,
            True,
        ],
        [
            [ParameterString(name="test1"), ParameterString(name="test2")],
            [ParameterString(name="test2"), ParameterString(name="test1")],
            False,
            True,
        ],
        [[ParameterString(name="test3")], [ParameterString(name="test1")], False, False],
        [{"foo": "bar"}, {"foo": "bar"}, False, True],
        [{"bar": "foo", "foo": "bar"}, {"foo": "bar", "bar": "foo"}, False, True],
        [{"bar": "bar"}, {"foo": "bar"}, False, False],
    ],
)
def test_equals(observed: List, expected: List, negative: bool, success: bool) -> None:
    equals = Equals(negative=negative)
    vr = equals.run(
        observed=observed,
        expected=expected,
        validation_name="some-validation",
    )
    assert vr.success == success


@pytest.mark.parametrize(
    "observed,expected,negative,success",
    [
        [["test1", "test2"], ["test1", "test2"], False, True],
        [["test1"], ["test1", "test2"], False, False],
        [["test1", "test2"], ["test1", "test2"], True, False],
        [["test1"], ["test1", "test2"], True, True],
        [["test1", "test2"], ["test2"], False, True],
        [["test1", "test2"], ["test3", "test4"], False, False],
        [
            [ParameterString(name="test1"), ParameterString(name="test2")],
            [ParameterString(name="test2"), ParameterString(name="test1")],
            False,
            True,
        ],
        [[ParameterString(name="test3")], [ParameterString(name="test1")], False, False],
    ],
)
def test_contains_str(
    observed: List[str], expected: List[str], negative: bool, success: bool
) -> None:
    contains = Contains(negative=negative)
    vr = contains.run(
        observed=observed,
        expected=expected,
        validation_name="some-validation",
    )
    assert vr.success == success
